import sqlite3
import click

class Database:

    def __init__(self, name):
        self._con = sqlite3.connect(f"{name}_lib.db")
        self._cur = self._con.cursor()
        self._cur.execute("CREATE TABLE IF NOT EXISTS Book(book_id INTEGER PRIMARY KEY AUTOINCREMENT, title VARCHAR(30), author VARCHAR(20), availability INTEGER)")
        self._cur.execute("CREATE TABLE IF NOT EXISTS Member(member_id INTEGER PRIMARY KEY AUTOINCREMENT, name VARCHAR(20), password VARCHAR(16))")
        self._cur.execute("CREATE TABLE IF NOT EXISTS Borrowed_books(member_id INTEGER, book_id INTEGER, FOREIGN KEY(member_id) REFERENCES Member(member_id), FOREIGN KEY(book_id) REFERENCES Book(book_id), PRIMARY KEY(member_id, book_id))")
        self._con.commit()

    def close_connection(self):
        self._con.close()

    def save_update_delete(self, query, params):
        self._cur = self._con.cursor()
        if params != None:
            self._cur.execute(query, params)
        else:
            self._cur.execute(query)
        self._con.commit()
        return True

    def read_data(self, query):
        self._cur = self._con.cursor()
        self._cur.execute(query)
        return self._cur.fetchall()

class Library:
    def __init__(self, name):
        self._name = name
        self._books = []
        self._members = []
        self._db = Database(self._name)

    def __repr__(self):
        return f"Library(Name = {self._name}, Books = {len(self._books)}, Members = {len(self._members)})"

    def get_name(self):
        return self._name

    def add_book(self, book):
        query = "INSERT INTO Book (title, author, availability) VALUES(?, ?, ?)"
        params = (book._title, book._author, book._a)
        if(self._db.save_update_delete(query, params)):
            book_id = self._db._cur.lastrowid
            book._book_id = book_id
            print(f"\nSuccessfully added: {book._title}\nBook ID: {book_id}")
        self._books.append(book)
        

    def register_member(self, member):
        query = f"INSERT INTO Member (name, password) VALUES (?, ?)"
        params = (member._name, member._password)
        if(self._db.save_update_delete(query, params)):
            member_id = self._db._cur.lastrowid
            member._member_id = member_id
            print(f"\nSuccessfully registered: {member._name}\nYour Member ID: {member_id} (YOU WILL NEED THIS ID TO LOGIN)")
        self._members.append(member)

    def load_all_books(self): #Load details of all books
        query = "SELECT * FROM Book"
        rows = self._db.read_data(query)
        print(f"\n{'Book ID':<10}{'Title':<30}{'Author':<20}{'Availability':<10}")
        print("-" * 72)
        for row in rows:
            book_id, title, author, availability = row
            print(f"{book_id:<10}{title:<30}{author:<20}{availability:<10}")
    
    def load_book(self, book_id): #Load details of a single book
        query = f"SELECT * FROM Book WHERE book_id = {book_id}"
        rows = self._db.read_data(query)
        for row in rows:
            return row

    def load_all_members(self):
        query = "SELECT * FROM Member"
        rows = self._db.read_data(query)
        for rows in row:
            print(rows)

    def load_member(self, member_id):
        query = f"SELECT * FROM Member WHERE member_id = {member_id}"
        rows = self._db.read_data(query)
        for row in rows:
            return row

    def login(self, member_id, password):
        query = f"SELECT * FROM Member WHERE member_id = {member_id} AND password = '{password}'"
        if self._db.read_data(query):
            return True
        else:
            return False

    def load_member_data(self, member):
        query = f"SELECT m.member_id, m.name, m.password, GROUP_CONCAT(b.title, ', ') AS borrowed_books FROM Member m LEFT JOIN Borrowed_books bb ON m.member_id = bb.member_id LEFT JOIN Book b ON bb.book_id = b.book_id WHERE m.member_id = {member.get_member_id()} GROUP BY m.member_id"
        rows = self._db.read_data(query)
        if rows:
            print("\nHere are your details:")
            print(f"\nMember ID      : {rows[0][0]}")
            print(f"Name           : {rows[0][1]}")
            print(f"Password       : {rows[0][2]}")
            print(f"Borrowed Books : {rows[0][3]}\n")
        else:
            print("An unexpected error has occured, please try again.")
            

    def check_borrowed_status(self, member_id, book_id):
        query = f"SELECT * FROM borrowed_books WHERE member_id = {member_id} AND book_id = {book_id}"
        rows = self._db.read_data(query)
        for row in rows:
            return row

    def load_borrowed_books(self, member_id):
        query = f"SELECT Borrowed_books.book_id, Book.title FROM Borrowed_books JOIN Book ON Borrowed_books.book_id = Book.book_id WHERE Borrowed_books.member_id = {member_id}"
        rows = self._db.read_data(query)
        if rows:
            borrowed_books = f"\n{'Book ID':<10} {'Title':<30}\n"
            line = "-" * 40
            borrowed_books += f"{line}\n\n"
            borrowed_books += "\n".join([f"{row[0]:<10} {row[1]}" for row in rows])
            return borrowed_books
        else:
            return False

class Book:
    def __init__(self, title, author, a, lib, book_id=None): #a = availability
        self._title = title
        self._author = author
        self._a = a
        self._book_id = book_id
        self._db = Database(lib.get_name())

    def __repr__(self):
        return f"Book(Book ID = {self._book_id}, Title = {self._title}, Author = {self._author}, Availability = {self._a})"

    def borrow_book(self):
        if self._a > 0:
            self._a -= 1
            print(f"\nSuccessfully borrowed: {self._title}")
            return True
        else:
            print(f"\n{self._title} currently not available.")
            return False

    def get_availability(self):
        return self._a

    def return_book(self):
        self._a += 1

class Member:
    def __init__(self, name, password, lib, member_id=None):
        self._name = name
        self._password = password
        self._member_id = member_id
        self._brwd_books = []
        self._db = Database(lib.get_name())

    def __repr__(self):
        return f"Member(Name = {self._name}, Borrowed Books = {self._brwd_books})"

    def get_member_id(self):
        return self._member_id

    def borrow_book(self, book):
        result = book.borrow_book()
        if not result:
            return

        query1 = "UPDATE Book SET availability = ? WHERE book_id = ?"
        params1 = (book.get_availability(), book._book_id)

        query2 = f"INSERT INTO Borrowed_books VALUES(?, ?)"
        params2 = (self._member_id, book._book_id)
        
        if (self._db.save_update_delete(query1, params1) and self._db.save_update_delete(query2, params2)):
            pass

        else:
            print("\nAn unknown error has occured. Please try again.")

        #self._brwd_books.append(book._title)

    def return_book(self, book):
        book.return_book()

        query1 = "UPDATE Book SET availability = ? WHERE book_id = ?"
        params1 = (book.get_availability(), book._book_id)

        query2 = "DELETE FROM Borrowed_books WHERE member_id = ? AND book_id = ?"
        params2 = (self._member_id, book._book_id)

        if (self._db.save_update_delete(query1, params1) and self._db.save_update_delete(query2, params2)):
            print(f"\nSuccessfully returned: {book._title}")

        #self._brwd_books.remove(book._book_id)

@click.group()
@click.pass_context
def library(ctx):
    lib = Library("BookInsider")
    member = None
    ctx.obj = [lib, member]

@library.command()
@click.argument('member_id')
@click.argument('password')
@click.pass_context
def member_login(ctx, member_id, password):
    lib = ctx.obj[0]
    member = ctx.obj[1]
    if lib.login(member_id, password):
        member_details = lib.load_member(member_id)
        member = Member(member_details[1], member_details[2], lib, member_details[0])
        ctx.obj[1] = member
        return member_details[1]
    else:
        return False

@library.command()
@click.argument('name')
@click.argument('password')
@click.pass_context
def member_register(ctx, name, password):
    lib = ctx.obj[0]
    new_member = Member(name, password, lib)
    lib.register_member(new_member)

@library.command()
@click.argument('title')
@click.argument('author')
@click.argument('availability')
@click.pass_context
def add_book(ctx, title, author, availability):
    lib = ctx.obj[0]
    new_book = Book(title, author, availability, lib)
    lib.add_book(new_book)

@library.command()
@click.pass_context
def load_all_books(ctx):
    ctx.obj[0].load_all_books()

@library.command()
@click.argument('member_id')
@click.argument('password')
@click.argument('book_id')
@click.pass_context
def borrow_book(ctx, member_id, password, book_id):
    lib = ctx.obj[0]
    member = ctx.obj[1]
    if lib.login(member_id, password): #Check for correct credentials
        member_details = lib.load_member(member_id)
        if not member_details:
            click.echo("\nAn error has occured with your profile. Please try again.")
            return
        member = Member(member_details[1], member_details[2], lib, member_details[0])
        ctx.obj[1] = member
        book_details = lib.load_book(book_id)
        if not book_details:
            click.echo("\nBook not found. Please check if the entered Book ID is correct.")
            return
        borrowed = lib.check_borrowed_status(member_details[0], book_id)
        if borrowed:
            click.echo(f"\nYou already borrowed the book you specified: {book_details[1]}")
            return
        book = Book(book_details[1], book_details[2], book_details[3], lib, book_details[0])
        member.borrow_book(book)
    else:
        click.echo("\nCredentials are invalid. Please check and try again.")

@library.command()
@click.argument('member_id')
@click.argument('password')
@click.argument('book_id')
@click.pass_context
def return_book(ctx, member_id, password, book_id):
    lib = ctx.obj[0]
    member = ctx.obj[1]
    if lib.login(member_id, password): #Check for correct credentials
        member_details = lib.load_member(member_id)
        if not member_details:
            click.echo("\nAn error has occured with your profile. Please try again.")
            return
        member = Member(member_details[1], member_details[2], lib, member_details[0])
        ctx.obj[1] = member
        book_details = lib.load_book(book_id)
        if not book_details:
            click.echo("\nBook not found. Please check if the entered Book ID is correct.")
            return
        borrowed = lib.check_borrowed_status(member_details[0], book_id)
        if not borrowed:
            click.echo(f"\nYou haven't borrowed the book you specified: {book_details[1]}")
            return

        book = Book(book_details[1], book_details[2], book_details[3], lib, book_details[0])
        member.return_book(book)
    else:
        click.echo("\nCredentials are invalid. Please check and try again.")

@library.command()
@click.argument('member_id')
@click.argument('password')
@click.pass_context
def load_borrowed_books(ctx, member_id, password):
    lib = ctx.obj[0]
    member = ctx.obj[1]
    if lib.login(member_id, password): #Check for correct credentials
        member_details = lib.load_member(member_id)
        if not member_details:
            click.echo("\nAn error has occured with your profile. Please try again.")
            return
        result = lib.load_borrowed_books(member_details[0])
        if not result:
            return False
        return result

    else:
        click.echo("\nCredentials are invalid. Please check and try again.")

@library.command()
@click.argument('member_id')
@click.argument('password')
@click.pass_context
def load_member_data(ctx, member_id, password):
    lib = ctx.obj[0]
    member = ctx.obj[1]
    if lib.login(member_id, password): #Check for correct credentials
        member_details = lib.load_member(member_id)
        if not member_details:
            click.echo("\nAn error has occured with your profile. Please try again.")
            return
        member = Member(member_details[1], member_details[2], lib, member_details[0])
        ctx.obj[1] = member
        member_data = lib.load_member_data(member)

    else:
        click.echo("\nCredentials are invalid. Please check and try again.")

@library.command()
def menu():
    while True:
        click.echo("\nLibrary Menu:")
        click.echo("\n1. Login")
        click.echo("2. Register")
        click.echo("3. Add a book")
        click.echo("4. List all books")
        click.echo("0. Exit")
        click.echo("\nTIP: REGISTER AND LOGIN TO BORROW BOOKS!")
        choice = click.prompt("\nEnter an option", type=str)
        ctx = click.get_current_context()

        if choice == "1":
            member_id = click.prompt("\nEnter your Member ID")
            password = click.prompt("Enter your password")
            
            name = ctx.invoke(member_login, member_id=member_id, password=password)
            if name == False:
                click.echo("\nInvalid credentials, please try again")
                continue
            click.echo(f"\nHello {name}, welcome back to the library!")

            while True:
                click.echo("\n1. Borrow a book")
                click.echo("2. Return a book")
                click.echo("3. List all borrowed books")
                click.echo("4. List all your data")
                click.echo("0. Back to main menu")
                sub_menu_pick = click.prompt("\nEnter an option", type=str)
            
                if sub_menu_pick == "1":
                    click.echo("\nHere are the list of available books\n")
                    ctx.invoke(load_all_books)
                    book_id = click.prompt("\nEnter the ID of the Book you want to borrow")
                    ctx.invoke(borrow_book, member_id = member_id, password=password, book_id=book_id)
                elif sub_menu_pick == "2":
                    click.echo("\nHere are the list of your borrowed books:\n")
                    result = ctx.invoke(load_borrowed_books, member_id=member_id, password=password)
                    if not result:
                        click.echo("\nYou haven't borrowed any books yet.\n")
                        continue
                    click.echo(f"{result}")
                    book_id = click.prompt("\nEnter the ID of the Book you want to return")
                    ctx.invoke(return_book, member_id = member_id, password=password, book_id=book_id)

                elif sub_menu_pick == "3":
                    result = ctx.invoke(load_borrowed_books, member_id=member_id, password=password)
                    if not result:
                        click.echo("\nYou haven't borrowed any books yet.\n")
                        continue
                    click.echo(f"{result}")

                elif sub_menu_pick == "4":
                    ctx.invoke(load_member_data, member_id=member_id, password=password)

                elif sub_menu_pick == "0":
                    break

                else:
                    click.echo("Invalid input, please chose from the list above.")
        
        elif choice == "2":
            name = click.prompt("Enter your name")
            password = click.prompt("Enter a new password")
            ctx.invoke(member_register, name=name, password=password)

        elif choice == "3":
            title = click.prompt("Enter the book Title")
            author = click.prompt("Enter it's Autor")
            availability = click.prompt("Enter it's availability")
            ctx.invoke(add_book, title=title, author=author, availability=availability)

        elif choice == "4":
            ctx.invoke(load_all_books)
            click.echo("\n0. Back\n")
            sub_choice = click.prompt(">")
            if sub_choice == "0":
                continue
            else:
                click.echo("\nInvalid input.")
                break

        elif choice == "0":
            break

        else:
            click.echo("\nInvalid input. Please chose from the list above.")

if __name__ == "__main__":

    library()
